function apply_rw(data::Array{Float64}, W::Matrix{Float64}, vcfg::Union{Nothing, Vector{Int32}}=nothing; id::Union{String, Nothing}=nothing, fs::Bool=false)
    nc = size(W,2)
    if isnothing(vcfg)
        vcfg = collect(1:nc)
    end
    
    if fs == false
        if id == "A653"
            rw1 = W[1, 1:nc] 
            rw = rw1
            data_r = data .* rw[vcfg]
            return (data_r, rw)
        else
            rw1 = W[1, 1:nc] 
            rw2 = W[2, 1:nc]
            rw = rw1 .* rw2
            data_r = data .* rw[vcfg]
            return (data_r, rw)
        end
    else
        rw1 = W[1, 1:nc] 
        rw2 = W[2, 1:nc]
        rw_s = [1.0 for i in 1:nc]
        if id in ["H105r001", "H105r002", "H105r005", "J303", "J303r003"]
            rw_s[flag_s[id]] .= -1.0
        end
        rw = rw1 .* rw2 .* rw_s
        data_r = data .* rw[vcfg]
        return (data_r, rw)
    end
end

function apply_rw(data::Vector{<:Array{Float64}}, W::Vector{Matrix{Float64}}, vcfg::Union{Nothing, Vector{Vector{Int32}}}=nothing; id::Union{String, Nothing}=nothing, fs::Bool=false)
    nc = size.(W, 2)
    if isnothing(vcfg)
        vcfg = [collect(1:nc[k]) for k=1:length(nc)]
    end

    if fs == false
        if id == "A653"
            rw1 = [W[k][1, 1:nc[k]] for k=1:length(W)]    
            rw = [rw1[k] for k =1:length(W)]
            data_r = [data[k] .* rw[k][vcfg[k]] for k=1:length(data)]
            return (data_r, rw)
        else
            rw1 = [W[k][1, 1:nc[k]] for k=1:length(W)]
            rw2 = [W[k][2, 1:nc[k]] for k=1:length(W)]
            rw = [rw1[k] .* rw2[k] for k =1:length(W)]
            data_r = [data[k] .* rw[k][vcfg[k]] for k=1:length(data)]
            return (data_r, rw)
        end
    else
        rw1 = [W[k][1, 1:nc[k]] for k=1:length(W)]
        rw2 = [W[k][2, 1:nc[k]] for k=1:length(W)]
        rw_s = [[1.0 for i in 1:nc[k]] for k=1:length(W)]
        if id == "H105"
            rw_s[1][flag_s["H105r001"]] .= -1.0
            rw_s[2][flag_s["H105r002"]] .= -1.0
        elseif id == "H105r005"
            rw_s[1][flag_s["H105r005"]] .= -1.0
        elseif id == "J303" || id == "J303r003"
            rw_s[1][flag_s["J303r003"]] .= -1.0
        end
        rw = [rw1[k] .* rw2[k] .* rw_s[k] for k =1:length(W)]
        data_r = [data[k] .* rw[k][vcfg[k]] for k=1:length(data)]
        return (data_r, rw)
    end
end

function check_corr_der(obs::Corr, derm::Vector{Corr}; new_version::Bool=false)   
    g1 = Vector{String}(undef, 0)
    g2 = Vector{String}(undef, 0)

    for d in derm
        if !new_version
            aux = [d.gamma[1], d.gamma[2]]
            push!(g1, aux[1][1:end-3])
            push!(g2, aux[2][1:end-3])
        else
            aux = [d.gamma[1], d.gamma[2]]
            push!(g1, aux[1][1:end])
            push!(g2, aux[2][1:end])
        end
    end

    h = copy(derm)
    push!(h, obs)
    
    
    if any(getfield.(h, :y0) .!= getfield(h[1], :y0))
        return false
    end
    for s in [:kappa, :mu]
        for k = 1:2
            if any(getindex.(getfield.(h, s), k) .!= getindex(getfield(h[1], s), k))
                return false
            end
        end
    end
    #gamma check
    if any(g1 .!= obs.gamma[1]) || any(g2 .!= obs.gamma[2])
        return false
    end

    return true
end

@doc raw"""
    corr_obs(cdata::CData; real::Bool=true, rw::Union{Array{Float64, 2}, Nothing}=nothing, L::Int64=1, info::Bool=false)

    corr_obs(cdata::Array{CData, 1}; real::Bool=true, rw::Union{Array{Array{Float64, 2}, 1}, Nothing}=nothing, L::Int64=1, info::Bool=false)

Creates a `Corr` struct with the given `CData` struct `cdata` (`read_mesons`) for a single replica.
An array of `CData` can be passed as argument for multiple replicas.

The flag `real` select the real or imaginary part of the correlator.
If `rw` is specified, the method applies reweighting. `rw` is passed as a matrix of Float64 (`read_ms1`)
The correlator can be normalized with the volume factor if `L` is fixed.

The flag `info` provides extra output that contains information about the primary observables. The function returns the primary observables ``<WO>`` and ``<W>``
(it returns the observable <O> if rw=nothing)

If you turn `flag_strange=true` and `rw` is passed, will reweight to undo negative sign of the strange quark determinant.

```@example
#Single replica
data = read_mesons(path, "G5", "G5")
rw = read_ms1(path_rw)
corr_pp = corr_obs.(data)
corr_pp_r = corr_obs.(data, rw=rw)

#Single replica + Info
data = read_mesons(path, "G5", "G5")
rw = read_ms1(path_rw)
corr_pp, O = corr_obs(data[1], info=true)
corr_pp_r, WO, W = corr_obs(data[1], rw=rw, info=true)

#Two replicas
data = read_mesons([path_r1, path_r2], "G5", "G5")
rw1 = read_ms1(path_rw1)
rw2 = read_ms1(path_rw2)

corr_pp = corr_obs.(data)
corr_pp_r = corr_obs.(data, rw=[rw1, rw2])
```
"""
function corr_obs(cdata::CData; real::Bool=true, rw::Union{Array{Float64, 2}, Nothing}=nothing, L::Int64=1, info::Bool=false, idm::Union{Vector{Int64},Nothing}=nothing, nms::Int64=Int64(maximum(cdata.vcfg)), flag_strange::Bool=false)

    real ? data = cdata.re_data ./ L^3 : data = cdata.im_data ./ L^3
    nt = size(data)[2]

    idm = isnothing(idm) ? Int64.(cdata.vcfg) : idm

    if isnothing(rw)
        # idm = isnothing(idm) ? collect(1:nms) : idm
        obs = [uwreal(data[:, x0], cdata.id, idm, nms) for x0 = 1:nt]
    else
        # idm = isnothing(idm) ? collect(1:nms) : idm
        data_r, W = apply_rw(data, rw, cdata.vcfg, id=cdata.id, fs=flag_strange)
        ow = [uwreal(data_r[:, x0], cdata.id, idm, nms) for x0 = 1:nt]
        W_obs = uwreal(W, cdata.id, idm, nms)
        obs = [ow[x0] / W_obs for x0 = 1:nt]
    end

    if info && !isnothing(rw)
        return (Corr(obs, cdata), ow, W_obs)
    elseif info && isnothing(rw)
        return (Corr(obs, cdata), obs)
    else
        return Corr(obs, cdata)
    end
end
#function corr_obs for R != 1
#TODO: vcfg with gaps
function corr_obs(cdata::Array{CData, 1}; real::Bool=true, rw::Union{Array{Array{Float64, 2}, 1}, Nothing}=nothing, L::Int64=1, info::Bool=false, replica::Union{Vector{Int64},Nothing}=nothing, idm::Union{Vector{Int64},Nothing}=nothing, nms::Union{Int64, Nothing}=nothing, flag_strange::Bool=false)
    nr = length(cdata)
    id = getfield.(cdata, :id)
    vcfg = getfield.(cdata, :vcfg)
    replica = isnothing(replica) ? Int64.(maximum.(vcfg)) : replica

    nms = isnothing(nms) ?  sum(replica) : nms

    nr = length(vcfg)
    if isnothing(idm)
        a = vcfg[1]
        for i in 2:nr
            a = [a; a[end] .+ vcfg[i]]
        end
        idm = Int64.(a)
    end
    
    if !all(id .== id[1])
        error("IDs are not equal")
    end

    real ? data = getfield.(cdata, :re_data) ./ L^3 : data = getfield.(cdata, :im_data) ./ L^3

    nt = size(data[1])[2]

    if isnothing(rw)
        tmp = data[1]
        [tmp = cat(tmp, data[k], dims=1) for k = 2:nr]
        obs = [uwreal(tmp[:, x0], id[1], replica, idm, nms) for x0 = 1:nt]
    else
        data_r, W = apply_rw(data, rw, vcfg, id=id[1], fs=flag_strange)

        tmp = data_r[1]
        tmp_W = W[1]
        [tmp = cat(tmp, data_r[k], dims=1) for k = 2:nr]
        [tmp_W = cat(tmp_W, W[k], dims=1) for k = 2:nr]

        ow = [uwreal(tmp[:, x0], id[1], replica, idm, nms) for x0 = 1:nt]
        W_obs = uwreal(tmp_W, id[1], replica, idm, nms)
        obs = [ow[x0] / W_obs for x0 = 1:nt]
    end
    if info && !isnothing(rw)
        return (Corr(obs, cdata), ow, W_obs)
    elseif info && isnothing(rw)
        return (Corr(obs, cdata), obs)
    else
        return Corr(obs, cdata)
    end
end   

@doc raw"""
    corr_obs_TSM(cdata_sl::CData, cdata_corr::CData; real::Bool=true, rw::Union{Array{Float64, 2}, Nothing}=nothing, L::Int64=1, info::Bool=false, idm_sl::Union{Vector{Int64},Nothing}=nothing, idm_corr::Union{Vector{Int64},Nothing}=nothing, nms::Union{Int64,Nothing}=nothing, flag_strange::Bool=false)
    corr_obs_TSM(cdata1::Array{CData, 1}, cdata2::Array{CData, 1}; real::Bool=true, rw::Union{Array{Array{Float64, 2}, 1}, Nothing}=nothing, L::Int64=1, info::Bool=false, replica_sl::Union{Vector{Int64},Nothing}=nothing, idm_sl::Union{Vector{Int64},Nothing}=nothing, idm_corr::Union{Vector{Int64},Nothing}=nothing, nms::Union{Int64, Nothing}=nothing, flag_strange::Bool=false)
   

Creates a `Corr` struct  for Truncated Solver Method (TSM) with the given `CData` structs `cdata_sl` and `cdata_corr` for a single replica.
Two arrays of `CData` can be passed as argument for multiple replicas.

The flag `real` select the real or imaginary part of the correlator.
If `rw` is specified, the method applies reweighting. `rw` is passed as a matrix of Float64 (`read_ms1`)
The correlator can be normalized with the volume factor if `L` is fixed.

The flag `info` provides extra output that contains information about the primary observables. The function returns the primary observables ``<WO>`` and ``<W>``
(it returns the observable <O> if rw=nothing)

The flags `idm_sl`, `idm_corr` and `nms` can be used if the observable is not measure in every configuration. In particular:
`idm_sl::Vector{Int64}` If given, label the configurations where the sloppy observables is measured. Not required if sloppy is full statistics.    
`idm_corr::Vector{Int64}` If given, label the configurations where the correction observables is measured. Not required if correction is full statistics.
`nms::Int64` Total number of measurements in the ensembles. By default, nms=maximum(cdata_sl.vcfg) it is assumed that the sloppy observable is full statistics, while the exact observable is not.
Typically, only `idm_corr` has to be passed, if the exact observable is not full statistics. 
By playing with these three flags you ensure that the Gamma method is still working even if the configurations between sloppy and correction do not match.
`replica_sl::Vector{Int64}` allows to set the length of each replica independently of `maximum.(vcfg_sl)`. Be aware that `replica_sl` must be equal or larger than `maximum(vcfg_sl)`

```@example
#Single replica
data_sloppy      = read_mesons(path_sl, "G5", "G5")
data_correction  = read_mesons_correction(path_corr, "G5", "G5")

rw = read_ms1(path_rw)
corr_pp = corr_obs_TSM.(data_sloppy, data_correction)
corr_pp_r = corr_obs_TSM.(data_sloppy, data_correction rw=rw)

#Single replica + Info
data_sloppy      = read_mesons(path_sl, "G5", "G5")
data_correction  = read_mesons_correction(path_corr, "G5", "G5")
rw = read_ms1(path_rw)
corr_pp, O = corr_obs_TSM(data_sloppy, data_correction, info=true)
corr_pp_r, WO, W = corr_obs(data_sloppy, data_correction,, rw=rw, info=true)

#Two replicas
data = read_mesons([path_r1, path_r2], "G5", "G5")
rw1 = read_ms1(path_rw1)
rw2 = read_ms1(path_rw2)

corr_pp = corr_obs.(data)
corr_pp_r = corr_obs.(data, rw=[rw1, rw2])
```
"""
function corr_obs_TSM(cdata_sl::CData, cdata_corr::CData; real::Bool=true, rw::Union{Array{Float64, 2}, Nothing}=nothing, L::Int64=1, info::Bool=false, idm_sl::Union{Vector{Int64},Nothing}=nothing, idm_corr::Union{Vector{Int64},Nothing}=nothing, nms::Union{Int64,Nothing}=nothing, flag_strange::Bool=false)
    # cdata1 is sloppy, cdata2 is correction 
    if cdata_sl.id != cdata_corr.id
        error("Error: cdata_sl id != cdata_corr id")
    end
    if cdata_sl.header != cdata_corr.header # Base.:(==) and Base.:(!=) are redifined in juobs_types.jl
        error("Error: cdata_sl header != cdata_corr header")
    end

    id = getfield(cdata_sl, :id)

    vcfg_sl   = getfield(cdata_sl, :vcfg)
    vcfg_corr = getfield(cdata_corr, :vcfg)
    nms = isnothing(nms) ? Int64(maximum(vcfg_sl)) : nms # assuming vcfg_sl >= vcfg_corr
    
    idm_sl = isnothing(idm_sl) ? Int64.(vcfg_sl) : idm_sl
    idm_corr = isnothing(idm_corr) ? Int64.(vcfg_corr) : idm_corr
    

    data1 = real ? cdata_sl.re_data   ./ L^3 : cdata_sl.im_data   ./ L^3
    data2 = real ? cdata_corr.re_data ./ L^3 : cdata_corr.im_data ./ L^3

    nt = size(data1, 2) # 1
    # added automatic idsm detection  
    # nms_vec_sl   = cdata_sl.vcfg
    # nms_vec_corr = cdata_corr.vcfg
 
    # delta_cnfg_sl   = nms_vec_sl[2] - nms_vec_sl[1]
    # delta_cnfg_corr = nms_vec_corr[2] - nms_vec_corr[1]
 
    # idm_sl   = isnothing(idm_sl) ? collect(1:delta_cnfg_sl:nms_vec_sl[end]) : idm_sl
    # idm_corr = isnothing(idm_corr) ? collect(1:delta_cnfg_corr:nms_vec_corr[end]) : idm_corr
    # end section to be tested 
    if isnothing(rw)
        obs1 = [uwreal(data1[:, x0], id, idm_sl, nms) for x0 = 1:nt]
        obs2 = [uwreal(data2[:, x0], id, idm_corr, nms) for x0 = 1:nt]
    else
        data1_r, W_sl = apply_rw(data1, rw, vcfg_sl, id=id, fs=flag_strange)
        data2_r, W_corr = apply_rw(data2, rw, vcfg_corr, id=id, fs=flag_strange)
        ow1 = [uwreal(data1_r[:, x0], id, idm_sl, nms) for x0 = 1:nt]
        ow2 = [uwreal(data2_r[:, x0], id, idm_corr, nms) for x0 = 1:nt]

        W_obs_sl   = uwreal(W_sl, id, idm_sl, nms)
        W_obs_corr = uwreal(W_corr, id, idm_corr, nms)

        obs1 = [ow1[x0] / W_obs_sl for x0 = 1:nt]
        obs2 = [ow2[x0] / W_obs_corr for x0 = 1:nt]
    end

    if info && !isnothing(rw)
        return (Corr(obs1 + obs2, cdata_sl), ow1, ow2, W_obs_sl)
    elseif info && isnothing(rw)
        return (Corr(obs1 + obs2, cdata_sl), obs1, obs2)
    else
        return Corr(obs1 + obs2, cdata_sl)
    end

end
function corr_obs_TSM(cdata1::Array{CData, 1}, cdata2::Array{CData, 1}; real::Bool=true, rw::Union{Array{Array{Float64, 2}, 1}, Nothing}=nothing, L::Int64=1, info::Bool=false, replica_sl::Union{Vector{Int64},Nothing}=nothing, idm_sl::Union{Vector{Int64},Nothing}=nothing, idm_corr::Union{Vector{Int64},Nothing}=nothing, nms::Union{Int64, Nothing}=nothing, flag_strange::Bool=false)
    if any(getfield.(cdata1, :id) .!= getfield.(cdata2, :id))
        error("Error: cdata1 id != cdata2 id")
    end
    if any(getfield.(cdata1, :header) .!= getfield.(cdata2, :header)) # Base.:(==) and Base.:(!=) are redifined in juobs_types.jl
        error("Error: cdata1 header != cdata2 header")
    end
    
    id = getfield.(cdata1, :id)
    if !all(id .== id[1])
        error("IDs are not equal")
    end

    vcfg_sl      = getfield.(cdata1, :vcfg)
    replica_sl   = isnothing(replica_sl) ? Int64.(maximum.(vcfg_sl)) : replica_sl
    vcfg_corr    = getfield.(cdata2, :vcfg)
    replica_corr = Int64.(maximum.(vcfg_corr))
    
    nms = isnothing(nms) ? sum(replica_sl) : nms # assuming vcfg_sl >= vcfg_corr

    nr = length(vcfg_sl)
    if isnothing(idm_sl)
        a = vcfg_sl[1]
        for i in 2:nr
            a = [a; a[end] .+ vcfg_sl[i]]
        end
        idm_sl = Int64.(a)
    end
    if isnothing(idm_corr)
        a = vcfg_corr[1]
        for i in 2:nr
            a = [a; a[end] .+ vcfg_corr[i]]
        end
        idm_corr = Int64.(a)
    end

    data1 = real ? getfield.(cdata1, :re_data) ./ L^3 : getfield.(cdata1, :im_data) ./ L^3
    data2 = real ? getfield.(cdata2, :re_data) ./ L^3 : getfield.(cdata2, :im_data) ./ L^3

    nt = size(data1[1], 2)
    if isnothing(rw)
        tmp1 = cat(data1..., dims=1)
        tmp2 = cat(data2..., dims=1)

        obs1 = [uwreal(tmp1[:, x0], id[1], replica_sl, idm_sl, nms) for x0 = 1:nt]
        obs2 = [uwreal(tmp2[:, x0], id[1], replica_sl, idm_corr, nms) for x0 = 1:nt]

    else
        data1_r, W = apply_rw(data1, rw, vcfg_sl, id=id[1], fs=flag_strange)
        data2_r, W = apply_rw(data2, rw, vcfg_corr, id=id[1], fs=flag_strange)

        tmp1 = cat(data1_r..., dims=1)
        tmp2 = cat(data2_r..., dims=1)
        tmp_W = cat(W..., dims=1)

        ow1 = [uwreal(tmp1[:, x0], id[1], replica_sl, idm_sl, nms) for x0 = 1:nt]
        ow2 = [uwreal(tmp2[:, x0], id[1], replica_sl, idm_corr, nms) for x0 = 1:nt]

        W_obs = uwreal(tmp_W, id[1], replica_sl, idm_sl, nms)

        obs1 = [ow1[x0] / W_obs for x0 = 1:nt]
        obs2 = [ow2[x0] / W_obs for x0 = 1:nt]

    end
    if info && !isnothing(rw)
        return (Corr(obs1 + obs2, cdata1), ow1, ow2, W_obs)
    elseif info && isnothing(rw)
        return (Corr(obs1 + obs2, cdata1), obs1, obs2)
    else
        return Corr(obs1 + obs2, cdata1)
    end
end     
@doc raw"""
    corr_sym(corrL::Corr, corrR::Corr, parity::Int64=1)

Computes the symmetrized correlator using the left correlador `corrL` and the right correlator `corrR`. The source position
of `corrR` must be `T - 1 - y0`, where `y0` is the source position of `corrL`. 

```@example
pp_sym = corr_sym(ppL, ppR, +1)
a0p_sym = corr_sym(a0pL, a0pR, -1)
```
"""
function corr_sym(corrL::Corr, corrR::Corr, parity::Int64=1)
    T = length(corrL.obs)
    sym = [:kappa, :mu, :gamma]
    if corrL.y0 != T - 1 - corrR.y0
        error("Corr: Parameter mismatch")
    end
    for s in sym
        if getfield(corrL, s) != getfield(corrR, s)
            error("Corr: Parameter mismatch")
        end
    end
    if abs(parity) != 1
        error("incorrect value of parity (+- 1)")
    end

    res = (corrL.obs[1:end] + parity * corrR.obs[end:-1:1]) / 2
    return Corr(res, corrL.kappa, corrL.mu, corrL.gamma, corrL.y0, corrL.theta1, corrL.theta2)
end

function corr_sym_E250(corr1::Corr, corr2::Corr, parity::Int64=1)
    aux = [corr2.obs[97:end]; corr2.obs[1:96]]
    corr2_sym = Corr(aux, corr2.kappa, corr2.mu, corr2.gamma, corr1.y0, corr2.theta1, corr2.theta2)

    corr = [corr1.obs[1:3]; (corr1.obs[4:98] .+ parity * corr2_sym.obs[192:-1:98]) / 2]

    return Corr(corr, corr1.kappa, corr1.mu, corr1.gamma, corr1.y0, corr1.theta1, corr1.theta2)
end

function corr_sym_D450(corr1::Corr, corr2::Corr, parity::Int64=1)
    aux = [corr2.obs[65:end]; corr2.obs[1:64]]
    corr2_sym = Corr(aux, corr2.kappa, corr2.mu, corr2.gamma, corr1.y0, corr2.theta1, corr2.theta2)

    corr = [corr1.obs[1:3]; (corr1.obs[4:66] .+ parity * corr2_sym.obs[128:-1:66]) / 2]

    return Corr(corr, corr1.kappa, corr1.mu, corr1.gamma, corr1.y0, corr1.theta1, corr1.theta2)
end

#TODO: VECTORIZE, uwreal?
@doc raw"""
    md_sea(a::uwreal, md::Vector{Matrix{Float64}}, ow::uwreal, w::Union{uwreal, Nothing}=nothing, ws::ADerrors.wspace=ADerrors.wsg, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    md_sea(a::uwreal, md::Vector{Matrix{Float64}}, ow::Array{uwreal}, w::Union{uwreal, Nothing}=nothing, ws::ADerrors.wspace=ADerrors.wsg, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

Computes the derivative of an observable A with respect to the sea quark masses.

``\frac{d <A>}{dm(sea)} = \sum_i \frac{\partial <A>}{\partial <O_i>}  \frac{d <O_i>}{d m(sea)}``


``\frac{d <O_i>}{dm(sea)} = <O_i> <\frac{\partial S}{\partial m}> - <O_i \frac{\partial S}{\partial m}>``

where ``O_i`` are primary observables. The function returns  ``\frac{\partial <A>}{\partial <O_i>}  \frac{d <O_i>}{d m(sea)}``, where
the primary observable ``O_i`` is specified as an input argument.

`ow` is a primary observable or a vector of primary observables that contains ``<OW>`` (``<O>`` if reweighting is not applied) and 
`w` is a primary obserable that contains the reweighting factors ``<W>`` (`w`=nothing if reweighting is not applied).

`md` is a vector that contains the derivative of the action ``S`` with respect
to the sea quark masses for each replica. `md[irep][irw, icfg]`

`md_sea` returns a tuple of uwreal observables ``(dA/dm_l, dA/dm_s)|_{sea}``, 
where ``m_l`` and ``m_s`` are the light and strange quark masses.

```@example
#Single replica
data = read_mesons(path, "G5", "G5")
md = read_md(path_md)
rw = read_ms1(path_rw)

corr_pp, wpp, w = corr_obs.(data[1], rw=rw, info=true)
m = meff(corr_pp[1], plat)
m_mdl, m_mds = md_sea(m, [md], wpp, w, ADerrors.wsg)
m_shifted = m + 2 * dml * m_mdl + dms * m_mds

#Two replicas
data = read_mesons([path_r1, path_r2], "G5", "G5")
md1 = read_md(path_md1)
md2 = read_md(path_md2)

corr_pp, pp = corr_obs(data[1], info=true)
m = meff(corr_pp[1], plat)
m_mdl, m_mds = md_sea(m, [md1, md2], pp, ADerrors.wsg)
m_shifted = m + 2 * dml * m_mdl + dms * m_mds
```
"""
function md_sea(a::uwreal, md::Vector{Matrix{Float64}}, ow::uwreal, w::Union{uwreal, Nothing}=nothing, ws::ADerrors.wspace=ADerrors.wsg, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    nid = neid(a)
    p = findall(t-> t==1, a.prop)

    if nid != 1
        @info("Error: neid > 1")
    end

    id = ws.map_nob[p]
    if !all(id .== id[1])
        @info("ids do not match")
    end
    id = ws.id2str[id[1]]
    
    ivrep = getfield.(ws.fluc[p], :ivrep)
    ivrep1 = fill(ivrep[1], length(ivrep))
    if !all(ivrep .== ivrep1)
        @info("ivreps do not match")
    end
    ivrep = ivrep[1]

    if length(md) != length(ivrep)
        @info("Nr obs != Nr md")
    end

    #md_aux as a Matrix + Automatic truncation
    md_aux = md[1][:, 1:ivrep[1]]
    for k = 2:length(md)
        md_aux = cat(md_aux, md[k][:, 1:ivrep[k]], dims=2)
    end
    
    nrw = size(md_aux, 1)
    isnothing(wpm) ? uwerr(ow) : uwerr(ow,wpm)
    ow_data = ow.mean .+ mchist(ow, id)

    if isnothing(w)
        der = derivative(a, ow)
        d1 = der != 0.0 ? der * (ow * uwreal(md_aux[1, :], id, ivrep) - uwreal(ow_data .* md_aux[1, :], id, ivrep)) : 0.0
        
        if nrw == 1
            return (d1, d1)
        elseif nrw == 2
            d2 = der != 0.0 ? der * (ow * uwreal(md_aux[2, :], id, ivrep) - uwreal(ow_data .* md_aux[2, :], id, ivrep)) : 0.0
            return (d1, d2)
        end
    else
        uwerr(w)
        der = derivative(a, ow) * w.mean
        w_data = w.mean .+ mchist(w, id)
        d1 = der != 0.0 ? der * (ow * uwreal(w_data .* md_aux[1, :], id, ivrep) / w^2 - uwreal(ow_data .* md_aux[1, :], id, ivrep) / w) : 0.0
        
        if nrw == 1
            return (d1, d1)
        elseif nrw == 2
            d2 = der != 0.0 ? der * (ow * uwreal(w_data .* md_aux[2, :], id, ivrep) / w^2 - uwreal(ow_data .* md_aux[2, :], id, ivrep) / w) : 0.0
            return (d1, d2)
        end
    end

    return nothing
end

function md_sea(a::uwreal, md::Vector{Matrix{Float64}}, ow::Array{uwreal}, w::Union{uwreal, Nothing}=nothing, ws::ADerrors.wspace=ADerrors.wsg, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    d = [md_sea(a, md, ow_, w, ws, wpm) for ow_ in ow]
    return (sum(getindex.(d, 1)), sum(getindex.(d, 2)))
end

@doc raw"""
    md_val(a::uwreal, obs::Corr, derm::Vector{Corr}; new_version::Bool=false, TSM::Bool=false)

Computes the derivative of an observable A with respect to the valence quark masses. 
If `new_version=true`, then the sign of the valence derivatives in `derm` is flipped 
(Julien's most updated code has the sign changed for the valence derivatives). Note 
also that for the new dat files version, no `_d1, _d2` is written by the reader, so 
you must select by hand (looking at the log files e.g.) which are you derivative 
correlators. In the log file, the derivative correlators are signaled by `seq_prop=some number`.

``\frac{d <A>}{dm(val)} = \sum_i \frac{\partial <A>}{\partial <O_i>}  \frac{d <O_i>}{d m(val)}``

``\frac{d <O_i>}{dm(val)} = <\frac{\partial O_i}{\partial m(val)}>``

where ``O_i`` are primary observables 

`md` is a vector that contains the derivative of the action ``S`` with respect
to the sea quark masses for each replica. `md[irep][irw, icfg]`

`md_val` returns a tuple of `uwreal` observables ``(dA/dm_1, dA/dm_2)|_{val}``, 
where ``m_1`` and ``m_2`` are the correlator masses.

```@example
data = read_mesons(path, "G5", "G5", legacy=true)
data_d1 = read_mesons(path, "G5_d1", "G5_d1", legacy=true)
data_d2 = read_mesons(path, "G5_d2", "G5_d2", legacy=true)

rw = read_ms1(path_rw)

corr_pp = corr_obs.(data, rw=rw)
corr_pp_d1 = corr_obs.(data_d1, rw=rw)
corr_pp_d2 = corr_obs.(data_d2, rw=rw)
derm = [[corr_pp_d1[k], corr_pp_d2[k]] for k = 1:length(pp_d1)]

m = meff(corr_pp[1], plat)
m_md1, m_md2 = md_val(m, corr_pp[1], derm[1])
m_shifted = m + dm1 * m_md1 + dm2 * m_md2
```
"""
function md_val(a::uwreal, obs::Corr, derm::Vector{Corr}; new_version::Bool=false)
    nid = neid(a)
    if nid != 1
        error("Error: neid > 1")
    end
    if length(derm) != 2
        error("Error: length derm != 2")
    end
    if !check_corr_der(obs, derm; new_version=new_version)
        error("Corr parameters does not match")
    end

    corr = getfield(obs, :obs)
    prop = getfield.(corr, :prop)

    if all(count.(prop) .== 1)
        der = [derivative(a, corr[k]) for k = 1:length(corr)]
    elseif all(count.(prop) .== 2)
        corr_der = getfield.(corr, :der)
        n = findall.(t-> t==1, prop)
        n = vcat(n'...)
        if all(n[:, 1] .== n[1, 1]) # find ow and w
            n_w = n[1, 1]
            n_ow = n[:, 2]
        else
            n_w = n[1, 2]
            n_ow = n[:, 1]
        end
        w_mean = 1 / getindex.(corr_der, n_ow)[1]
        #= Extra information (ow, w)
        ws = ADerrors.wsg
        ow_mean = - getindex.(corr_der, n_w) .* w_mean^2
        w_data = w_mean .+ ws.fluc[n_w].delta
        fluc_ow = getfield.(ws.fluc[n_ow], :delta)
        fluc_ow = vcat(fluc_ow'...)
        ow_data = ow_mean .+ ws.fluc[n_ow]
        =#
        der = length(a.der) >= maximum(n_ow) ? a.der[n_ow] * w_mean : 0.0
    else
        return nothing
    end
    derm1, derm2 = derm
    if !new_version
        return (sum(der .* derm1.obs), sum(der .* derm2.obs))
    else
        return (-sum(der .* derm1.obs), -sum(der .* derm2.obs))
    end
end

"""  
    plat_av(obs::Vector{uwreal}, plat::Vector{Int64}, [W::VecOrMat{Float64},] wpm::Union{Dict{Int64},Vector{Float{64}},Dict{String,Vector{Float{64}},Nothing}=nothing} )

It computes plateau average of `obs` in the range `plat`. Effectively this perform a fit to a constant. 
The field `W` is optional. When absent, `W` is a `Vector{Float64}` constructed from the error of `obs`. 
When `W` is a `Vector{Float64}` it performs an uncorrelated fit, when `W` is a `Matrix{Float64}`, it perfom a correlated fit

# Potential side effects

When `W` is absent, the function acts with `ADerror.uwerr` on `obs` in order to compute the errors.
This is the reason why a `wpm` may be given as input.

"""
function plat_av(obs::Vector{uwreal}, plat::Vector{Int64}, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(obs) : [uwerr(obs_aux, wpm) for obs_aux in obs]
    w = 1 ./ err.(obs)[plat[1]:plat[2]].^2
    av = sum(w .* obs[plat[1]:plat[2]]) / sum(w)
    return av 
end

plat_av(obs::Vector{uwreal}, plat::Vector{Int64}, W::Vector{Float64},wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing) = sum(W .* obs[plat[1]:plat[2]]) / sum(W)
function plat_av(obs::Vector{uwreal}, plat::Vector{Int64}, W::Matrix{Float64},wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
  o = obs[plat[1]:plat[2]]
  return sum([o[i]*W[i,j]*o[j] for i in axes(W,1), j in axes(W,2)])/sum([o[i]*W[i,j] for i in axes(W,1), j in axes(W,2)])
end
function model_av(fun::Function, y::Vector{uwreal}, guess::Float64; 
    tm::Vector{Int64}, tM::Vector{Int64}, k::Int64, 
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing) 
    
    pval = Array{Float64,1}()
    p_1 = Array{uwreal,1}()
    TIC = Array{Float64,1}()

    isnothing(wpm) ? uwerr.(y) : [uwerr(y[i], wpm) for i in 1:length(y)]
    for i in tm
        for j in tM
            if i < j
                x = collect(i:j)
                y_aux = y[i:j]
                try 
                    up, chi2, chi_exp, pval_i = fit_alg(fun,x,y_aux,k,guess,wpm=wpm)
                    push!(pval, pval_i)
                    push!(TIC, chi2 - 2*chi_exp)
                    push!(p_1, up[1])
                catch e 
                end
            end
        end
    end

    TIC = TIC .- minimum(TIC)
    weight = exp.(-0.5 .* TIC) ./ sum(exp.(-0.5 .* TIC))
    p_av = sum(p_1 .* weight)
    syst2 = sum(p_1 .^ 2 .* weight) - p_av ^ 2
    return p_av, syst2, p_1, weight, pval
end

function model_av(fun::Vector{Function}, y::Vector{uwreal}, guess::Float64; 
    tm::Vector{Vector{Int64}}, tM::Vector{Vector{Int64}}, k::Vector{Int64}, 
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing) 
    
    pval = Array{Float64,1}()
    p_1 = Array{uwreal,1}()
    TIC = Array{Float64,1}()

    isnothing(wpm) ? uwerr.(y) : [uwerr(y[i], wpm) for i in 1:length(y)]
    for ind in 1:length(fun)
        f = fun[ind]
        for i in tm[ind]
            for j in tM[ind]
                if i < j
                    x = collect(i:j)
                    y_aux = y[i:j]
                    try 
                        up, chi2, chi_exp, pval_i = fit_alg(f,x,y_aux,k[ind],guess,wpm=wpm)
                        push!(pval, pval_i)
                        push!(TIC, chi2 - 2*chi_exp)
                        push!(p_1, up[1])
                    catch e 
                    end
                end
            end
        end
    end

    TIC = TIC .- minimum(TIC)
    weight = exp.(-0.5 .* TIC) ./ sum(exp.(-0.5 .* TIC))
    p_av = sum(p_1 .* weight)
    syst2 = sum(p_1 .^ 2 .* weight) - p_av ^ 2
    return p_av, syst2, p_1, weight, pval
end

@doc raw"""
    bayesian_av(fun::Function, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k::Int64, pl::Bool, data::Bool; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    bayesian_av(fun1::Function, fun2::Function, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k1::Int64, k2::Int64, pl::Bool, data::Bool; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    bayesian_av(fun::Array{Function}, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k::Array{Int64}, pl::Bool, data::Bool; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

Computes bayesian average of data. For a given fit function, it explores choices of fit intervals, assigning each of them a weight. The function saves the `first` fit parameter of your function, and then it does the weighted average of it and assigns a systematic. See https://arxiv.org/abs/2008.01069 

The function takes as input the fit intervals to explore. 

`tmin_array` is an array of integers with the lower bounds on the fit intervals to explore, ***ordered from lower to higher***.

`tmax_array` is an array of integers with the upper bounds on the fit intervals to explore, ***ordered from lower to higher***.

`k` is the number of parameters of the fit function to use.

You can also use as input two fit functions, and two values of `k`, one for each function. Then, for each fit interval choice, the function explores the two fit functions. This means that for each fit interval choice you get two results: one for the first fit funcction, and another for the second. You can also use a vector of functions and a vector of k (numer of parameters of each funtion) to apply the bayesian averaging method to multiple functions.

The method returns two objects: first, the weighted average as an uwreal object, with mean value and statistichal error. The second object returned is the systematic error coming from the fit interval variation. If `data` is `true`, then returns 4 objects: weighted average, systematic error, a vector with the results of the fit for each fit interval choice, and a vector with the weights associated to each fit.

```@example
@.fun(x,p) = p[1] * x ^0
k = 1
tmin_array = [10,11,12,13,14,15]
tmax_array = [80,81,82,83,84,85]
(average, systematics, data, weights) = bayesian_av(fun,x,tmin_array,tmax_array,k,pl=true,data=true)

@.fun1(x,p) = p[1] * x ^0
@.fun2(x,p) = p[1] + p[2] * exp( - p[3] * (x))
k1 = 1
k2 = 3
tmin_array = [10,11,12,13,14,15]
tmax_array = [80,81,82,83,84,85]
(average, systematics) = bayesian_av(fun1,fun2,x,tmin_array,tmax_array,k1,k2)

@.fun1(x,p) = p[1] * x ^0
@.fun2(x,p) = p[1] + p[2] * exp( - p[3] * (x))
k1 = 1
k2 = 3
tmin_array = [10,11,12,13,14,15]
tmax_array = [80,81,82,83,84,85]
(average, systematics) = bayesian_av([fun1,fun2],x,tmin_array,tmax_array,[k1,k2])
```
"""
function bayesian_av(fun::Function, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k::Int64; pl::Bool=false, data::Bool=false,
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing, path_plt::Union{String,Nothing}=nothing, 
    plt_title::Union{Nothing,String}=nothing, label::Union{Nothing, LaTeXString}=nothing)
    
    weight_model = Array{Float64,1}()
    AIC = Array{Float64,1}()
    chi2chi2exp = Array{Float64,1}()
    p1 = Array{uwreal,1}()
    mods = Array{String,1}()
    Pval = Array{Float64,1}()

    if tmax_array[end] > length(y)
        error("Error: upper bound for the fits is bigger than last data point")
    end

    total = length(y)
    isnothing(wpm) ? uwerr.(y) : [uwerr(y[i],wpm) for i in 1:length(y)]
    
    for INDEX in tmin_array ## vary tmin
        for j in tmax_array ## vary tmax
            try
                x = [i for i in INDEX+1:1:j]	
	            yy = y[INDEX+1:1:j]
	            Ncut = total - length(x)
                dy = err.(yy)
                W = 1 ./ dy .^2

                p00 = [0.5 for i in 1:1:k]
                chisq = gen_chisq(fun,x,dy)
                fit = curve_fit(fun,x,value.(yy),W,p00)
                isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),yy) : (up,chi_exp) = fit_error(chisq,coef(fit),yy,wpm)
                isnothing(wpm) ? uwerr(up[1]) : uwerr(up[1],wpm)
                chi2 = sum(fit.resid.^2) # * dof(fit) / chi_exp
                Q = pvalue(chisq, sum(fit.resid.^2), value.(up), yy, wpm=wpm, W=W, nmc=10000)
                # if chi2 / dof(fit) > 5.  || abs(value(up[1])) > 2.  || err(up[1])/value(up[1])*100 >= 10.
                    # error()
                # end
                # println(up[1], " ", up[2], " ", up[3])
                push!(Pval, Q)

                # push!(AIC, chi2 + 2*k + 2*Ncut)   # AIC criteria
                push!(AIC, chi2 - 2*chi_exp) # TIC criteria

                push!(chi2chi2exp, chi2 / dof(fit)) #substitute chi_exp with  dof(fit) for chi^2/dof
                push!(p1, up[1])
                push!(mods,string("[", INDEX+1, ",", j, "]"))
                # println("chi2_vs_exp " , chi2 / dof(fit),  " chi2/dof ", sum(fit.resid.^2)/dof(fit))
            catch 
                @warn string(":/ Negative window for error propagation at tmin = ", INDEX, ", tmax = ", j, "; skipping that point")
            end
        end
    end
    
    # compute all weights
    offset = minimum(AIC)
    AIC = AIC .- offset
    weight_model = exp.(-0.5 .* AIC)
    weight_model = weight_model ./ sum(weight_model)
    
    # sort weights and discard 5% tail 
    idxW = sortperm(weight_model, rev=true)
    cumulative_w = cumsum(weight_model[idxW])
    # println("acceptance in bayeasian_av set to 0.99, should be 0.95")
    idxcumw = findfirst(x->x>=0.95, cumulative_w)
    # println(length(idxW) - idxcumw)
    idxW = sort(idxW[1:idxcumw])
    
    # take only the accepted 
    AIC = AIC[idxW]
    Pval = Pval[idxW]
    chi2chi2exp  = chi2chi2exp[idxW]
    mods = mods[idxW]
    p1 = p1[idxW]
    weight_model = weight_model[idxW]
    weight_model ./= sum(weight_model)


    
    p1_mean = sum(p1 .* weight_model)
    isnothing(wpm) ? uwerr(p1_mean) : uwerr(p1_mean, wpm) 
    # println(sum(p1 .^ 2 .* weight_model))
    # println((sum(p1 .* weight_model)) ^ 2)
    systematic_err =  0.0 #sqrt(sum(p1 .^ 2 .* weight_model) - (sum(p1 .* weight_model)) ^ 2) 
    if pl   
        fig = figure(figsize=(10,7.5))

        subplots_adjust(hspace=0.1) 
        subplot(411)    
        if !isnothing(plt_title)
            #title(plt_title)
        end
        ax1 = gca()                
        x = 1:length(p1)
        y = value.(p1)
        dy = err.(p1)
        v = value(p1_mean)
        e = err(p1_mean)

        fill_between(x, v-e, v+e, color="royalblue", alpha=0.2)
        errorbar(x, y, dy, fmt="^", mfc="none", color="navy", capsize=2)
        setp(ax1.get_xticklabels(),visible=false) # Disable x tick labels
        isnothing(label) ?  ylabel(L"$p_1$") : ylabel(label)
    
        subplot(412)
        ax2=gca()
        bar(x, weight_model, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5)
        setp(ax2.get_xticklabels(),visible=false) # Disable x tick labels
        # errorbar(mods, weight_model, 0*dy, color="green")
        ylabel(L"$W$")

        subplot(413)
        ax3=gca()
        setp(ax3.get_xticklabels(),visible=false) # Disable x tick labels

        bar(x, Pval, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5 )
        ylabel(L"$p-value$")
        
        
        subplot(414)
        bar(x, chi2chi2exp, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5 )
        # ylabel(L"$\chi^2/\chi^2_{\mathrm{exp}}$")
        ylabel(L"$\chi^2/\mathrm{d.o.f.}$")
        xticks(x, mods, rotation=45)
        xlabel(L"$\mathrm{Models} \ [t_{\mathrm{min}}/a,t_{\mathrm{max}}/a]$")
        tight_layout()
        display(fig)
        if !isnothing(path_plt)
            #tt = plt_title *".pdf" 
            savefig(path_plt)
        end
        close()
    end

    if !data                   
        return (p1_mean, systematic_err)                                         
    else               
        FitStorage = Dict(
        "W"        => weight_model,
        "AIC"      => AIC .+ offset,
        "pval"     => Pval, 
        "chivsexp" => chi2chi2exp,
        "mods"     => mods
        )
        return (p1_mean, systematic_err, p1, FitStorage)    
    end 

end

function bayesian_av(fun1::Function, fun2::Function, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k1::Int64, k2::Int64; pl::Bool=false, data::Bool=false, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    weight_model = Array{Float64,1}()
    AIC = Array{Float64,1}()
    chi2chi2exp = Array{Float64,1}()
    p1 = Array{uwreal,1}()
    mods = Array{String,1}()

    if tmax_array[end] > length(y)
        error("Error: upper bound for the fits is bigger than last data point")
    end

    total = length(y)
    isnothing(wpm) ? uwerr.(y) : for i in 1:length(y) uwerr(y[i],wpm) end
    
    for INDEX in tmin_array ## vary tmin
        for j in tmax_array ## vary tmax
            try
                x = [i for i in INDEX+1:1:j]	
	            yy = y[INDEX+1:1:j]
	            Ncut = total - length(x)
                dy = err.(yy)
                W = 1 ./ dy .^2

                p00 = [0.5 for i in 1:1:k1]
                chisq = gen_chisq(fun1,x,dy)
                fit = curve_fit(fun1,x,value.(yy),W,p00)
                isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),yy) : (up,chi_exp) = fit_error(chisq,coef(fit),yy,wpm)
                isnothing(wpm) ? uwerr(up[1]) : uwerr(up[1],wpm)
                chi2 = sum(fit.resid.^2) * dof(fit) / chi_exp
                push!(AIC, chi2 + 2*k1 + 2*Ncut)
                push!(chi2chi2exp, chi2 / dof(fit))
                push!(p1, up[1])

                push!(mods,string("[", INDEX+1, ",", j, "]"))

                p00 = [0.5 for i in 1:1:k2]
                chisq = gen_chisq(fun2,x,dy)
                fit = curve_fit(fun2,x,value.(yy),W,p00)
                isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),yy) : (up,chi_exp) = fit_error(chisq,coef(fit),yy,wpm)
                uwerr(up[1],wpm)
                chi2 = sum(fit.resid.^2) * dof(fit) / chi_exp
                push!(AIC, chi2 + 2*k2 + 2*Ncut)
                push!(chi2chi2exp, chi2 / dof(fit))
                push!(p1, up[1])
                println(up[end])
                push!(mods,string("[", INDEX+1, ",", j, "]"))

            catch e
                @warn string(":/ Negative window for error propagation at tmin = ", INDEX, ", tmax = ", j, "; skipping that point")
            end
        end
    end

    offset = minimum(AIC)
    AIC = AIC .- offset
    weight_model = exp.(-0.5 .* AIC)
    
    p1_mean = sum(p1 .* weight_model)/sum(weight_model) ; isnothing(wpm) ? uwerr(p1_mean) : uwerr(p1_mean,wpm) 
    weight_model = weight_model ./ sum(weight_model)
    systematic_err = sqrt(sum(p1 .^ 2 .* weight_model) - (sum(p1 .* weight_model)) ^ 2) 

    if pl                       
        x = 1:length(p1)
        y = value.(p1)
        dy = err.(p1)
        v = value(p1_mean)
        e = err(p1_mean)
        
        figure()
        fill_between(1:length(p1), v-e, v+e, color="green", alpha=0.75)
        errorbar(mods, y, dy, fmt="x", color="black")
        ylabel(L"$p_1$")
        xlabel(L"model")
    
        display(gcf())

        figure()
        errorbar(mods, weight_model, 0*dy, color="green")
        ylabel(L"$weight$")
        xlabel(L"model")
    
        display(gcf())
    end

    if !data                   
        return (p1_mean, systematic_err)                                         
    else               
        return (p1_mean, systematic_err, p1, weight_model)    
    end 

end

function bayesian_av(fun::Array{Function}, y::Array{uwreal}, tmin_array::Array{Int64}, tmax_array::Array{Int64}, k::Array{Int64}; pl::Bool=false,
    data::Bool=false, wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing, 
    path_plt::Union{String,Nothing}=nothing, plt_title::Union{Nothing,String}=nothing)

    weight_model = Array{Float64,1}()
    AIC          = Array{Float64,1}()
    chi2chi2exp  = Array{Float64,1}()
    p1           = Array{uwreal,1}()
    mods         = Array{String,1}()
    Pval         = Array{Float64,1}()

    if tmax_array[end] > length(y)
        error("Error: upper bound for the fits is bigger than last data point")
    end

    total = length(y)
    isnothing(wpm) ? uwerr.(y) : [uwerr(y[i],wpm) for i in 1:length(y)] 
    
    for INDEX in tmin_array ## vary tmin
        for j in tmax_array ## vary tmax
            try
                x = [i for i in INDEX+1:1:j]	
	            yy = y[INDEX+1:1:j]
	            Ncut = total - length(x)
                dy = err.(yy)
                W = 1 ./ dy .^2

                for indice in 1:length(fun)
                    p00 = [0.5 for i in 1:1:k[indice]]
                    chisq = gen_chisq(fun[indice],x,dy)
                    fit = curve_fit(fun[indice],x,value.(yy),W,p00)
                    isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),yy) : (up,chi_exp) = fit_error(chisq,coef(fit),yy,wpm)
                    isnothing(wpm) ? uwerr(up[1]) : uwerr(up[1],wpm)
                    chi2 = sum(fit.resid.^2) * dof(fit) / chi_exp
                    Q = pvalue(chisq, sum(fit.resid.^2), value.(up), yy, wpm=wpm, W=W, nmc=10000)
                    if chi2 /  dof(fit) < 5.
                        # if err(up[1])/value(up[1])*100 >= 10.
                        # error()
                        # end
                        push!(Pval, Q)
                        push!(AIC, chi2 + 2*k[indice] + 2*Ncut)
                        push!(chi2chi2exp, chi2 / dof(fit))
                        push!(p1, up[1])
                        push!(mods,string("[", INDEX+1, ",", j, "]"))
                    end
                end

            catch e
                @warn string(":/ Negative window for error propagation at tmin = ", INDEX, ", tmax = ", j, "; skipping that point")
            end
        end
    end
    # compute all weights
    offset = minimum(AIC)
    AIC = AIC .- offset
    weight_model = exp.(-0.5 .* AIC)
    weight_model = weight_model ./ sum(weight_model)
    
    # sort weights and discard 5% tail 
    idxW = sortperm(weight_model, rev=true)
    cumulative_w = cumsum(weight_model[idxW])
    idxcumw = findfirst(x->x>=0.95, cumulative_w)
    println(length(idxW) - idxcumw)
    idxW = sort(idxW[1:idxcumw])
    
    # take only the accepted 
    AIC = AIC[idxW]
    Pval = Pval[idxW]
    chi2chi2exp  = chi2chi2exp[idxW]
    mods = mods[idxW]
    p1 = p1[idxW]
    weight_model = weight_model[idxW]
    weight_model /= sum(weight_model)
    
    p1_mean = sum(p1 .* weight_model) 
    isnothing(wpm) ? uwerr(p1_mean) : uwerr(p1_mean,wpm) 
    systematic_err = sqrt(sum(p1 .^ 2 .* weight_model) - (sum(p1 .* weight_model)) ^ 2) 

    if pl   
        subplots_adjust(hspace=0.1) 
        subplot(411)    
        if !isnothing(plt_title)
            title(plt_title)
        end
        ax1 = gca()                
        x = 1:length(p1)
        y = value.(p1)
        dy = err.(p1)
        v = value(p1_mean)
        e = err(p1_mean)

        fill_between(x, v-e, v+e, color="royalblue", alpha=0.2)
        errorbar(x, y, dy, fmt="^", mfc="none", color="navy", capsize=2)
        setp(ax1.get_xticklabels(),visible=false) # Disable x tick labels
        ylabel(L"$p_1$")
    
        subplot(412)
        ax2=gca()
        bar(x, weight_model, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5)
        setp(ax2.get_xticklabels(),visible=false) # Disable x tick labels
        # errorbar(mods, weight_model, 0*dy, color="green")
        ylabel(L"$W$")

        subplot(413)
        ax3=gca()
        setp(ax3.get_xticklabels(),visible=false) # Disable x tick labels

        bar(x, Pval, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5 )
        ylabel(L"$p\ value$")
        
        
        subplot(414)
        bar(x, chi2chi2exp, alpha=0.4, color="forestgreen", edgecolor="darkgreen", linewidth=1.5 )
        ylabel(L"$\chi^2/\chi^2_{\mathrm{exp}}$")
        xticks(x, mods, rotation=45)
        xlabel(L"$Models$")

        display(gcf())
        if !isnothing(path_plt)
            tt = plt_title *".pdf" 
            savefig(joinpath(path_plt, tt))
        end
        close()
    end
    

    if !data                   
        return (p1_mean, systematic_err)                                         
    else               
        FitStorage = Dict(
        "W"        => weight_model,
        "AIC"      => AIC .+ offset,
        "pval"     => Pval, 
        "chivsexp" => chi2chi2exp,
        "mods"     => mods
        )
        return (p1_mean, systematic_err, p1, FitStorage)
    end 

end

function lin_fit(x::Vector{<:Real}, v::Vector{Float64}, e::Vector{Float64})
    sig2 = e .* e
    S = sum(1 ./ sig2)
    Sx = sum(x ./ sig2)
    Sy = sum(v ./ sig2)
    Sxy = sum(v .* x ./ sig2)
    Sxx = sum(x .* x ./sig2)
    delta = S * Sxx - Sx*Sx
    par = [Sxx*Sy-Sx*Sxy, S*Sxy-Sx*Sy] ./delta
    #C = [[Sxx/delta, -Sx/delta], [-Sx/delta,  S/delta]]
    return par
end
@doc raw"""
    lin_fit(x::Vector{<:Real}, y::Vector{uwreal})

Computes a linear fit of uwreal data points y. This method return uwreal fit parameters and chisqexpected.

```@example
fitp, csqexp = lin_fit(phi2, m2)
m2_phys = fitp[1] + fitp[2] * phi2_phys
```
"""
function lin_fit(x::Vector{<:Real}, y::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(y) : [uwerr(yaux, wpm) for yaux in y]
    par = lin_fit(x, value.(y), err.(y))
    chisq(p, d) = sum((d .- p[1] .- p[2].*x).^2 ./ err.(y) .^2)
    (fitp, csqexp) = fit_error(chisq, par, y)
    for i = 1:length(fitp)
        isnothing(wpm) ? uwerr(fitp[i]) : uwerr(fitp[i], wpm)
        print("\n Fit parameter: ", i, ": ")
        details(fitp[i])
    end
    println("Chisq / chiexp: ", chisq(par, y), " / ", csqexp, " (dof: ", length(x)-length(par),")")
    return (fitp, csqexp)
end

@doc raw"""
    x_lin_fit(par::Vector{uwreal}, y::Union{uwreal, Float64})

Computes the results of a linear interpolation/extrapolation in the x axis
"""
x_lin_fit(par::Vector{uwreal}, y::Union{uwreal, Float64}) = (y - par[1]) / par[2]
@doc raw"""
    y_lin_fit(par::Vector{uwreal}, y::Union{uwreal, Float64})

Computes the results of a linear interpolation/extrapolation in the y axis
"""
y_lin_fit(par::Vector{uwreal}, x::Union{uwreal, Float64}) = par[1] + par[2] * x


@doc """
    get_chi(model::Function,x::Vector,par,data,W::VecOrMat{Float64})

It computes the chi square of a fit. `model` is assumed to be of the type f(xdata,par). 
If `W::Vector` the chisquare is computed for an uncorrelated fit
If `W::Matrix` the chisquare is computed for a correlated fit
"""
function get_chi(model::Function,x::Vector,par,data,W::Array{Float64,2})
  return sum((data.-model(x,par))' * W * (data.-model(x,par)))
end

function get_chi(model::Function,x::Vector,par,data,W::Vector{Float64})
  return sum((data.-model(x,par)).^2 .* W)
end

@doc raw"""
    fit_routine(model::Function, xdata::Array{<:Real}, ydata::Array{uwreal}, npar::Int64; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}}=Dict{Int64,Vector{Float64}}(),corr::Bool = true, W::VecOrMat{Float64}=Vector{Float64}(),guess::Vector{Float64} = fill(0.5,npar),logfile);

It executes a fit of the `ydata` with the fit function `model`.

# parameters
  - `npar::Int64`: number of fit parameters.
  - `wpm`: Windows parameter that ADerrors uses to computes the errors. This is used both when computing the error of the data and when computing the covariance matrix through `ADerrors.cov`
  - `corr::Bool`: if `true` (default) it exectues the correlated fit, this flag is overridden if `W` is passed as a `Vector`. When `corr=true`, and `W` is not given, it computes the covarinace matrix through `ADerrors.cov`, then it is inverted and symmetrized. 
  - `W::VecOrMat{Float64}`: Weight matrix/vector. When given, the flag `corr` is overridden. If `W` is a `Vector` the fit is perfomed uncorrelated,otherwise is correlated. 
  - `guess::Vector{Float64}`: Initial guess used in the fit routine.
  - `logfile`: handle to a logfile. If `nothing`, no log will be written. It can be `IO` file or a custom log structure that has an overloading for `print()` and `println()`.


# Returns
It returns a NamedTuple with names:
  - `:par`:    fit parameter as `Vector{uwreal}`
  - `:chi2`:   chisquare,
  - `:chiexp`: chisquare expected 
  - `:pval`:   pvalue 
"""
function fit_routine(model::Function,
  xdata::Array{<:Real},
  ydata::Array{uwreal}, 
  npar::Int64; 
  wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}}=Dict{Int64,Vector{Float64}}(),
  corr::Bool = true, 
  W::VecOrMat{Float64}=Vector{Float64}(),
  guess::Vector{Float64} = fill(0.5,npar),
  logfile = nothing,);

  nobs = length(ydata)
  if length(xdata)!=nobs
    error("xdata and ydata must be the of same size")
  end

  [uwerr(y, wpm) for y in ydata]
  if length(W) ==0
    if corr 
      C = ADerrors.cov(ydata,wpm)
      W = LinearAlgebra.pinv(W);
      W = (W'+W)*0.5
    else
      W = [1/y.err^2 for y in ydata]
    end
  end

  fit = LsqFit.curve_fit(model,xdata,ADerrors.value.(ydata),W,guess)

  chisq(p,d) = get_chi(model,xdata,p,d,W)
  chi2 = sum(fit.resid.^2)
  par = fit_error(chisq,LsqFit.coef(fit),ydata,wpm,W,false)

  chiexp = ADerrors.chiexp(chisq,LsqFit.coef(fit),ydata,wpm,W=W)

  pval = juobs.pvalue(chisq,chi2,LsqFit.coef(fit),ydata,W)

  if !isnothing(logfile)

    println(logfile, "juobs.fit_routine output:") 
    println(logfile, "\t\tFit routine in [$(xdata[1]),...,$(xdata[2])]")
    println(logfile, "\t\tparameters :", par...)
    println(logfile, "\t\tchi2:       ", chi2)
    println(logfile, "\t\tchiexp:     ", chiexp)
    println(logfile, "\t\tpvalue:     ", pval)
  end

  return (par=par,chi2=chi2,chiexp=chiexp,pval=pval)
end

#= function fit_routine(model::Vector{Function}, xdata::Vector{Array{Float64, N}} where N, ydata::Vector{Array{uwreal, N}} where N, param::Int64; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing,
    correlated_fit::Bool=false)

    if !(length(model) == length(xdata) == length(ydata))
        error("Dimension mismatch")
    end
    N = length(model)

    dat = ydata[1]
    idx = Vector{Vector{Int64}}(undef, N)
    e = Vector{Vector{Float64}}(undef, N)
    j = 1
    for i = 1:N
        if isnothing(wpm)
            uwerr.(ydata[i])
        else
            # println(yaux for yaux in ydata[i])
            for yaux in ydata[i]
                [uwerr(yaux[k], wpm) for k in 1:length(yaux)]
            end
            # [[uwerr(yaux[k], wpm) for k in eachindex(yaux)] for yaux in ydata[i]]
        end
        
        e[i] = err.(ydata[i])
        if i > 1 
            dat = vcat(dat, ydata[i])
        end
        stp = j + length(ydata[i]) - 1
        idx[i] = collect(j:stp)
        j = stp + 1
    end
    if !correlated_fit
        chisq = (par, dat) -> sum([sum((dat[idx[i]] .- model[i](xdata[i], par)).^2 ./e[i].^2) for i=1:N])
    else
        inv_cov_tot = juobs.inv_covar(vcat(ydata...))  #[inv_covar_multi_id(dat[idx[ii]]) for ii in 1:N]
        function chisq(par, dat)
            res = Vector{uwreal}(undef,0)
            for ii in 1:N
                res =  vcat(res, dat[idx[ii]] .- model[ii](xdata[ii],par))
            end 
            chi = reshape(res, 1,:) * inv_cov_tot*res
            return chi[1]
        end
    end

    min_fun(t) = chisq(t, value.(dat))
    p = fill(0.5, param)
    #println(chisq(p,dat))
    sol = optimize(min_fun, p, LBFGS()) 
    #println(chisq(sol.minimizer,dat))
    
    if !correlated_fit
        @info("Uncorrelated fit ")
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(chisq, sol.minimizer, dat) : fit_error(chisq, sol.minimizer, dat, wpm)
    else
        @info("Correlated fit ")

        #inv_cov_merge = zeros(size(inv_cov_tot[1],1)*N, size(inv_cov_tot[1],1)*N )
        #count = 1
        #for k in 1:N
        #    dim_tmp = size(inv_cov_tot[k], 1)
        #    inv_cov_merge[count:dim_tmp+count-1, count:dim_tmp+count-1] = inv_cov_tot[k]
        #    count +=dim_tmp
        #end
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(chisq, sol.minimizer, dat, W=inv_cov_tot) : fit_error(chisq, sol.minimizer, dat, wpm, W=inv_cov_tot)
    end

    println("Chisq / chiexp: ", min_fun(sol.minimizer), " / ", chi2_exp, " (dof: ", length(dat) - param,")")
    chis2_corrected = (length(dat) - param) * min_fun(sol.minimizer) / chi2_exp
    println("Chisq corrected: ", chis2_corrected)
    
    @info("Returns: params, chi2, chi2_exp\n")
    return upar, min_fun(sol.minimizer)  ,  chi2_exp
end

function fit_routine(model::Function, xdata::Array{uwreal}, ydata::Array{uwreal}, param::Int64=3; 
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing, covar::Bool=false)
    
    Nalpha = size(xdata, 2) # number of x-variables
    Ndata = size(xdata, 1) # number of datapoints
    if isnothing(wpm)
        uwerr.(ydata)
        uwerr.(xdata)
    else
        [uwerr(yaux, wpm) for yaux in ydata] 
        [uwerr(xaux, wpm) for xaux in xdata]
    end

    yval = value.(ydata)
    yer = err.(ydata)
    xval = value.(xdata)
    xer = err.(xdata)

    dat = Vector{Float64}(undef, Ndata * (Nalpha+1))
    ddat = Vector{Float64}(undef, Ndata * (Nalpha+1))
    data = Vector{uwreal}(undef, Ndata * (Nalpha+1)) 

    for i = 1:Nalpha
        dat[(i-1)*Ndata+1:i*Ndata] = xval[:, i]
        ddat[(i-1)*Ndata+1:i*Ndata] = xer[:, i]
        data[(i-1)*Ndata+1:i*Ndata] = xdata[:, i]
    end
    dat[Nalpha*Ndata+1:end] = yval
    ddat[Nalpha*Ndata+1:end] = yer
    data[Nalpha*Ndata+1:end] = ydata

    # Guess
    fit = curve_fit(model, xval, yval, 1.0 ./ yer.^2, fill(0.5, param))

    # Generate chi2 + solver
    if covar
        aux = Vector{Vector{uwreal}}(undef, Ndata)
        for k = 1:Ndata
            aux[k] = Vector{uwreal}(undef, Nalpha+1)
            for i = 1:Nalpha
                aux[k][i] = xdata[k, i]
            end
            aux[k][Nalpha+1] = ydata[k]
        end

        C = isnothing(wpm) ? [ADerrors.cov(aux[k]) for k = 1:Ndata] : [ADerrors.cov(aux[k], wpm) for k = 1:Ndata]
        chisq_full_cov(p, d) = get_chi2_cov(model, d, C, p, Nalpha)
        min_fun_cov(t) = chisq_full_cov(t, dat)
        sol = optimize(min_fun_cov, vcat(fit.param, dat[1:Nalpha*Ndata]), method=LBFGS())
        
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(chisq_full_cov, sol.minimizer, data) : fit_error(chisq_full_cov, sol.minimizer, data, wpm)
        println("Chisq / chiexp: ", min_fun_cov(sol.minimizer), " / ", chi2_exp, " (dof: ", length(ydata) - param,")")
        chis2_corrected = (length(ydata) - param) * min_fun_cov(sol.minimizer) / chi2_exp
        println("Chisq corrected: ", chis2_corrected)
        return upar, min_fun_cov(sol.minimizer) / chi2_exp, chi2_exp

    else
        chisq_full(p, d) = get_chi2(model, d, ddat, p, Nalpha)
        min_fun(t) = chisq_full(t, dat)
        sol = optimize(min_fun, vcat(fit.param, dat[1:Nalpha*Ndata]), method=LBFGS())
        
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(chisq_full, sol.minimizer, data) : fit_error(chisq_full, sol.minimizer, data, wpm)
        println("Chisq / chiexp: ", min_fun(sol.minimizer), " / ", chi2_exp, " (dof: ", length(ydata) - param,")")
        chis2_corrected = (length(ydata) - param) * min_fun(sol.minimizer) / chi2_exp
        println("Chisq corrected: ", chis2_corrected)
        return upar, min_fun(sol.minimizer) / chi2_exp, chi2_exp
    end

    #### chisq_full, min_fun out of conditional ->
    #### COMPILER WARNING ** incremental compilation may be fatally broken for this module **

    # Info
    #for i = 1:length(upar)
    #    isnothing(wpm) ? uwerr(upar[i]) : uwerr(upar[i], wpm)
    #    print("\n Fit parameter: ", i, ": ")
    #    details(upar[i])
    #end
    
end

function global_fit_routine(models::Vector{Function}, xdata::Vector{Array{Float64, 1}} , ydata::Vector{Array{uwreal, 1}}, n_par::Int64; guess_param::Vector{Float64}=fill(0.5, n_par), log::Bool=true,
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing,
    inv_cov::Union{Matrix{Float64}, Nothing}=nothing)

    if !(length(models) == length(ydata))
        error("Dimension mismatch")
    end
    N_sets = length(models)             # number of datasets
    x_flat = xdata[1]                   # unrolling xdata 
    y_flat = ydata[1]                   # unrolling ydata

    chunk_idx = Vector{Vector{Int64}}(undef, N_sets)
    j = 1
    for i = 1:N_sets
        isnothing(wpm) ? uwerr.(ydata[i]) :  [uwerr(yaux, wpm) for yaux in ydata[i]]
        if i > 1 
            y_flat = vcat(y_flat, ydata[i])
            x_flat =vcat(x_flat, xdata[i])
        end
        stp = j + length(ydata[i]) - 1
        chunk_idx[i] = collect(j:stp)
        j = stp + 1
    end
    function flat_model(x_flat::Vector{Float64}, param::Vector{Float64})
        fx = similar(x_flat)
        for ii in 1:N_sets
            mod = models[ii]
            fx[chunk_idx[ii]] = mod(x_flat[chunk_idx[ii]], param)
        end
        return fx
    end
    function flat_chisq(param, data)
        if isnothing(inv_cov)
            chi = 0.0
            for ii in 1:N_sets
                mod = models[ii]
                chi += sum((data[chunk_idx[ii]] .- mod(x_flat[chunk_idx[ii]], param)).^2 ./ err.(y_flat[chunk_idx[ii]]).^2 )        
            end
            return chi
        else 
            # take pieces of covariance and build correlated chi2 piece by piec 
            # chi =0.0
            # for ii in 1:N_sets
                # mod = models[ii]
                # cov_inv = juovs.inv_covar(ydata[ii])
                # res = ydata[ii] .- mod(xdata[ii], param)
                # chi += (reshape(res, 1,:) * cov_inv * res)[1]
            # end
            # return chi
            # take whole covariance and multiply for whole vector of dat-mod(x,param)
            res = Vector{uwreal}(undef, 0)
            for ii in 1:N_sets
                mod = models[ii]
                res =vcat(res, data[chunk_idx[ii]] .- mod(x_flat[chunk_idx[ii]], param))
            end
            chi = reshape(res, 1,:) * inv_cov * res  
            return chi[1]
        end
    end

    if isnothing(inv_cov)
        @info("Uncorrelated fit")
        fit = curve_fit(flat_model, x_flat, value.(y_flat), 1.0 ./ err.(y_flat).^2, guess_param )
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(flat_chisq, coef(fit) , y_flat) : fit_error(flat_chisq, coef(fit) , y_flat, wpm)
    else
        @info("Correlated fit")
        fit = curve_fit(flat_model, x_flat, value.(y_flat), inv_cov, guess_param )
        (upar, chi2_exp) = isnothing(wpm) ? fit_error(flat_chisq, coef(fit) , y_flat, W=inv_cov) : fit_error(flat_chisq, coef(fit) , y_flat, wpm, W=inv_cov)
    end
    println("chi2 from fit residue: ", sum(fit.resid.^2))
    chisq = flat_chisq(coef(fit), y_flat)  
    if log
        println("Converged param: ", fit.converged)
        println("Chisq / chiexp: ", chisq, " / ", chi2_exp, " (dof: ", dof(fit),")")
        chis2_corrected = (dof(fit)) * chisq / chi2_exp
        println("Chisq corrected: ", chis2_corrected)
    end
    return upar, value(chisq), chi2_exp, dof(fit)
end


function fit_alg(f::Function, x::Union{Vector{Int64}, Vector{Float64}, Matrix{Float64}}, y::Vector{uwreal}, 
    n::Int64, guess::Union{Float64, Vector{Float64}, Nothing}=nothing; 
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    
    isnothing(wpm) ? uwerr.(y) : [uwerr(y[i], wpm) for i in 1:length(y)]
    W = 1 ./ err.(y) .^ 2
    chisq = fit_defs(f,x,W)
    #lb = [-Inf for i in 1:n]
	#ub = [+Inf for i in 1:n]
    if guess == nothing
        p0 = [.5 for i in 1:n]
    else
        p0 = [guess; [1. for i in 1:n-length(guess)]]
        #lb[1] = .9 * guess  
        #ub[1] = 1.1 * guess
    end
    fit = curve_fit(f,x,value.(y),W,p0)#,lower=lb,upper=ub)
    chi2 = sum(fit.resid .^ 2)
    isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,coef(fit),y) : (up,chi_exp) = fit_error(chisq,coef(fit),y,wpm)
    isnothing(wpm) ? pval = pvalue(chisq,chi2,value.(up),y) : pval = pvalue(chisq,chi2,value.(up),y,wpm=wpm)
    return up, chi2, chi_exp, pval
end

function fit_alg(model::Function,x::Array{Float64},y::Array{uwreal},param::Int64,W::Matrix{Float64};
    guess::Union{Float64, Vector{Float64}, Nothing}=nothing,
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)

    if guess == nothing
        p00 = [.5 for i in 1:param]
    else
        p00 = [guess; [1. for i in 1:param-length(guess)]]
        #lb[1] = .9 * guess  
        #ub[1] = 1.1 * guess
    end

    chisq_corr(par,dat) = juobs.gen_chisq_correlated(model, x, W, par, dat)
    fit = curve_fit(model,x,value.(y),W,p00)
    up, chi_exp = fit_error(chisq_corr, coef(fit), y, wpm, W=W)
    uwerr.(up)
    chi2 = sum(fit.resid .^ 2)
    pval = pvalue(chisq_corr, sum(fit.resid .^ 2), value.(up), y, W)
    doff = dof(fit)

    return up, chi2, chi_exp, pval, doff
end


function fit_alg(f::Vector{Function}, x::Vector{Matrix{Float64}}, y::Vector{Vector{uwreal}}, 
    n::Int64, guess::Union{Float64, Vector{Float64}, Nothing}=nothing; 
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    
    y_n = y[1]
    x_n = x[1]
    for i in 2:length(x)
        y_n = vcat(y_n, y[i])
        x_n = vcat(x_n, x[i])
    end
    idx = [collect(1:size(x[i],1)) for i in 1:length(x)]
    [idx[i] = idx[i] .+ idx[i-1][end] for i in 2:length(idx)]

    isnothing(wpm) ? [uwerr.(y[i]) for i in 1:length(y)] : [[uwerr(y[i][j], wpm) for j in 1:length(y[i])] for i in 1:length(y)]
    W = [1 ./ err.(y[i]) .^ 2 for i in 1:length(y)]

    chisq = (par, y_n) -> sum([sum((y_n[idx[i]] .- f[i](x[i], par)) .^ 2 .* W[i]) for i in 1:length(x)])
    min_fun(t) = chisq(t, value.(y_n))
    if guess == nothing
        p0 = [.5 for i in 1:n]
    else
        p0 = [guess; [1. for i in 1:n-length(guess)]]
    end
    sol = optimize(min_fun, p0, Optim.Options(g_tol=1e-8, iterations=1000000)) 
    chi2 = min_fun(sol.minimizer)
    isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,sol.minimizer,y_n) : (up,chi_exp) = fit_error(chisq,sol.minimizer,y_n,wpm)
    isnothing(wpm) ? pval = pvalue(chisq,chi2,value.(up),y_n) : pval = pvalue(chisq,chi2,value.(up),y_n,wpm=wpm)
    return up, chi2, chi_exp, pval
end

##CHECK THIS ONE, NOT WORKING GOOD
function fit_alg(f::Vector{Function}, x::Vector{Matrix{Float64}}, y::Vector{Vector{uwreal}}, W::Vector{Matrix{Float64}}, 
    n::Int64, guess::Union{Float64, Vector{Float64}, Nothing}=nothing; 
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    
    y_n = y[1]
    x_n = x[1]
    for i in 2:length(x)
        y_n = vcat(y_n, y[i])
        x_n = vcat(x_n, x[i])
    end
    idx = [collect(1:size(x[i],1)) for i in 1:length(x)]
    [idx[i] = idx[i] .+ idx[i-1][end] for i in 2:length(idx)]

    isnothing(wpm) ? [uwerr.(y[i]) for i in 1:length(y)] : [[uwerr(y[i][j], wpm) for j in 1:length(y[i])] for i in 1:length(y)]

    chisq = (par, y_n) -> sum([juobs.gen_chisq_correlated(f[i], x[i], W[i], par, y_n[i]) for i in 1:length(x)])

    min_fun(t) = chisq(t, value.(y_n))
    if guess == nothing
        p0 = [.5 for i in 1:n]
    else
        p0 = [guess; [1. for i in 1:n-length(guess)]]
    end
    sol = optimize(min_fun, p0, Optim.Options(g_tol=1e-8, iterations=1000000)) 
    chi2 = min_fun(sol.minimizer)

    W_aux = Matrix{Float64}(undef, length(y_n), length(y_n))
    for i in 1:length(y_n)
        for j in 1:length(y_n)
            if i <= length(y[1])
                if j <= length(y[1])
                    W_aux[i,j] = W[1][i,j]
                else
                    W_aux[i,j] = .0
                end
            else
                if j <= length(y[1])
                    W_aux[i,j] = .0
                else
                    W_aux[i,j] = W[2][i-length(y[1]),j-length(y[1])]
                end
            end
        end
    end

    isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,sol.minimizer,y_n,W=W_aux) : (up,chi_exp) = fit_error(chisq,sol.minimizer,y_n,wpm,W=W_aux)
    isnothing(wpm) ? pval = pvalue(chisq,chi2,value.(up),y_n,W_aux) : pval = pvalue(chisq,chi2,value.(up),y_n,wpm=wpm,W_aux)
    return up, chi2, chi_exp, pval
end

function fit_alg(f::Vector{Function}, x::Vector{Matrix{Float64}}, y::Vector{Vector{uwreal}}, 
    n::Int64, lb::Vector{Float64}, ub::Vector{Float64}, guess::Union{Float64, Vector{Float64}, Nothing}=nothing; 
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    
    y_n = y[1]
    x_n = x[1]
    for i in 2:length(x)
        y_n = vcat(y_n, y[i])
        x_n = vcat(x_n, x[i])
    end
    idx = [collect(1:size(x[i],1)) for i in 1:length(x)]
    [idx[i] = idx[i] .+ idx[i-1][end] for i in 2:length(idx)]

    isnothing(wpm) ? [uwerr.(y[i]) for i in 1:length(y)] : [[uwerr(y[i][j], wpm) for j in 1:length(y[i])] for i in 1:length(y)]
    W = [1 ./ err.(y[i]) .^ 2 for i in 1:length(y)]

    chisq = (par, y_n) -> sum([sum((y_n[idx[i]] .- f[i](x[i], par)) .^ 2 .* W[i]) for i in 1:length(x)])
    min_fun(t) = chisq(t, value.(y_n))
    if guess == nothing
        p0 = [.5 for i in 1:n]
    else
        p0 = [guess; [1. for i in 1:n-length(guess)]]
    end
    sol = optimize(min_fun, lb, ub, p0, Fminbox(NelderMead()), Optim.Options(g_tol=1e-8, iterations=1000000))
    chi2 = min_fun(sol.minimizer)
    isnothing(wpm) ? (up,chi_exp) = fit_error(chisq,sol.minimizer,y_n) : (up,chi_exp) = fit_error(chisq,sol.minimizer,y_n,wpm)
    isnothing(wpm) ? pval = pvalue(chisq,chi2,value.(up),y_n) : pval = pvalue(chisq,chi2,value.(up),y_n,wpm=wpm)
    return up, chi2, chi_exp, pval
end

function fit_defs(f::Function,x,W) ## uncorrelated fit
	chisq(p,d) = sum((d .- f(x,p)).^2 .* W)
	return chisq
end

function gen_chisq(f::Function, x::Array{<:Real}, err::Vector{Float64}) #constrained
    chisq(par, dat) = sum((dat .- f(x,par)).^2 ./err.^2)
    return chisq
end
function gen_chisq_correlated(f::Function, x::Array{<:Real}, covar_inv::Matrix, par, dat )
    chi2 = 0.0
    #delta = dat - ffit_routine(x, par)
    #chi2 = (reshape(delta, 1,:) * covar_inv * delta )[1]
    for i in eachindex(dat)
        for j in eachindex(dat)
            #println(f(x,par)[i])
            #println(f(x,par)[j])
            #println(f(x[j,:],par))
            #println(f(x[i,:],par))
            chi2 = chi2 +  (dat[i] - f(x,par)[i]) * covar_inv[i,j] * (dat[j] - f(x,par)[j])
        end
    end
    #chisq(par, dat) = sum(
    #    transpose((dat .- f(x,par))) * covar_inv * (dat .- f(x,par))
    #) 
    return chi2
end

function get_chi2(f::Function, data, ddata, par, Nalpha) #full
    chi2 = 0.0

    Ndata = div(length(data), Nalpha+1)
    Npar = length(par) - Ndata * Nalpha
    p = par[1:Npar]

    for k = 1:Ndata
        xx = [par[Npar + k + (i-1)*Ndata] for i = 1:Nalpha]
        Cinv = zeros(Nalpha+1, Nalpha+1)
        [Cinv[i, i] = 1 / ddata[k + (i-1)*Ndata]^2 for i = 1:Nalpha+1]

        xx = [par[Npar + k + (i-1)*Ndata] for i = 1:Nalpha]
        delta = [data[k + (i-1)*Ndata] - xx[i] for i = 1:Nalpha]
        yy = f(xx', p)
        push!(delta, data[k + Nalpha*Ndata] - yy[1])

        chi2 += delta' * Cinv * delta
    end
    return chi2
end

function get_chi2_cov(f::Function, data, C, par, Nalpha) # full + cov
    chi2 = 0.0

    Ndata = div(length(data), Nalpha+1)
    Npar = length(par) - Ndata * Nalpha
    p = par[1:Npar]

    for k = 1:Ndata
        if det(C[k]) / prod(diag(C[k])) > 1e-6
            Cinv = inv(C[k])
        else
            Cinv = zeros(Nalpha+1, Nalpha+1)
            [Cinv[i, i] = 1 / C[k][i, i] for i = 1:Nalpha+1]
        end

        xx = [par[Npar + k + (i-1)*Ndata] for i = 1:Nalpha]
        delta = [data[k + (i-1)*Ndata] - xx[i] for i = 1:Nalpha]
        yy = f(xx', p)
        push!(delta, data[k + Nalpha*Ndata] - yy[1])
        
        chi2 += delta' * Cinv * delta 
    end
    return chi2
end

function get_covar(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(obs) :  [uwerr(yaux, wpm) for yaux in obs]
    ens = vcat(ensembles.(obs)...)
    tmp_fluc = mchist.(obs, ens)
    fluc = mapreduce(permutedims, vcat, tmp_fluc)' 
    covar = fluc' * fluc ./ size(fluc)[1]^2
    return covar
end
function get_covar_multi_id(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    isnothing(wpm) ? uwerr.(obs) :  [uwerr(yaux, wpm) for yaux in obs]
    id_set = getindex.(ensembles.(obs), 2) # hardcoded assuming that ens label is always the second entry in each ensembles(obs[i])
    
    cov_tot = zeros(length(id_set), length(id_set))
    count = 1
    for (k,id) in enumerate(unique(id_set))
        idx       = findall(x-> x==id, id_set )
        fluc_aux  = mchist.(obs[idx], fill(id, length(idx)))
        fluc      = mapreduce(permutedims, vcat, fluc_aux)'
        covar_set = fluc' * fluc ./ size(fluc)[1]^2
        dim_tmp   = size(covar_set,1)
        cov_tot[count:dim_tmp+count-1, count:dim_tmp+count-1] = covar_set
        count +=dim_tmp
    end
    return cov_tot
end
function inv_covar(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    cov = get_covar(obs, wpm=wpm)
    cov_inv = inv(cov)
    return (cov_inv' + cov_inv) / 2.
end
function inv_covar_multi_id(obs::Vector{uwreal}; wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing}=nothing)
    cov = get_covar_multi_id(obs, wpm=wpm)
    cov_inv = inv(cov)
    return (cov_inv' + cov_inv) / 2.
end =#

@doc raw"""
    pvalue(chisq::Function,
    chi2::Float64,
    xp::Vector{Float64}, 
    data::Vector{uwreal},
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing} = Dict{Int64,Vector{Float64}}();
    W::Union{Vector{Float64},Array{Float64,2}} = Vector{Float64}(),
    nmc::Int64 = 5000)

Computes the p-value of a previously done fit, using as input the `\chi^2` observed from the fit, the fit parameters and the fitted data. 
The p-value for a given `\chi^2` is the probability of, given the data you have, finding such a `\chi^2` or worse from a fit, and still
have the data well described by the fit function. `nmc` is the number of MC samples used to estimate the p-value integral, default is 5000.
By now it only works with a vector for weights (containing the diagonal of W)

```@example
function fit_defs(f::Function,x,W)
    chisq(p,d)=sum((d-f(x,p)).^2 .*W)
    return chisq
end

@.fun(x,p) = p[1] + p[2] * x
chisq = fit_defs(fun, x, 1.0 ./ err.(y) .^ 2)
fit = curve_fit(fun, x, value.(y), 1.0 ./ err.(y) .^ 2, [0.5, 0.5])
(up, chiexp) = fit_error(chisq, coef(fit), y)

wpm = Dict{Int64,Vector{Float64}}()
wpm[1] = [-1.0,-1.0,4-0,-1.0]
Q = pvalue(chisq, chi2, value.(up), y, wpm; W = 1.0 ./ err.(y) .^ 2, nmc=10000)
#Q = pvalue(chisq, chi2, value.(up), y; W = 1.0 ./ err.(y) .^ 2, nmc=10000)
#Q = pvalue(chisq, chi2, value.(up), y)
```
"""
function pvalue(chisq::Function,
    chi2::Float64,
    xp::Vector{Float64}, 
    data::Vector{uwreal},
    W::Vector{Float64}=Vector{Float64}();
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing} =  Dict{Int64,Vector{Float64}}(),
    nmc::Int64 = 5000,
    C::AbstractMatrix{Float64} = ADerrors.cov(data,wpm))

    n = length(xp)   # Number of fit parameters
    m = length(data) # Number of data
    if (m-n<0)
      error("more parameters than data")
    end

    xav = zeros(Float64, n+m)
    for i in 1:n
        xav[i] = xp[i]
    end
    for i in n+1:n+m
        xav[i] = data[i-n].mean
    end
    ccsq(x::Vector) = chisq(view(x, 1:n), view(x, n+1:n+m)) 
    if (n+m < 4)
        cfg = ForwardDiff.HessianConfig(ccsq, xav, ADerrors.Chunk{1}());
    else
        cfg = ForwardDiff.HessianConfig(ccsq, xav, ADerrors.Chunk{4}());
    end
        
    hess = Array{Float64}(undef, n+m, n+m)
    ForwardDiff.hessian!(hess, ccsq, xav, cfg)
    
    if length(W) ==0 
      W = zeros(Float64, m)
      for i in 1:m
        if (data[i].err == 0.0)
          uwerr(data[i], wpm)
          if (data[i].err == 0.0)
            error("Zero error in fit data")
          end
        end
        W[i] = 1.0 / data[i].err^2
      end
    end

    n = size(hess, 1) - m
    hm = view(hess, 1:n, n+1:n+m)
    sm = Array{Float64, 2}(undef, n, m)
    for i in 1:n, j in 1:m
        sm[i,j] = hm[i,j] / sqrt.(W[j])
    end
    maux = sm * sm'
    hi   = LinearAlgebra.pinv(maux)
    Px   = -hm' * hi * hm

    for i in 1:m
        Px[i,i] = W[i] + Px[i,i]
    end

    nu = sqrt(C) * Px * sqrt(C)
        
    N = length(nu[1,:])
    z = randn(N, nmc)

    eig = abs.(eigvals(nu))
    eps = 1e-14 * maximum(eig)
    eig = eig .* (eig .> eps)

    aux = eig' * (z .^ 2)
    Q = 1.0 - juobs.mean(aux .< chi2)

    x = chi2 .- eig[2:end]' * (z[2:end,:].^2)
    x = x / eig[1]
    #dQ = juobs.mean((x .> 0) .* exp.(-x * 0.5) * 0.5 ./ sqrt.(abs.(x)))
    #dQ = err(cse)/value(cse) * dQ
    return Q
end

function pvalue(chisq::Function,
    chi2::Float64,
    xp::Vector{Float64}, 
    data::Vector{uwreal},
    W::Array{Float64,2};
    wpm::Union{Dict{Int64,Vector{Float64}},Dict{String,Vector{Float64}}, Nothing} =  Dict{Int64,Vector{Float64}}(),
    nmc::Int64 = 5000,
    C::AbstractMatrix{Float64} = ADerrors.cov(data,wpm))

    n = length(xp)   # Number of fit parameters
    m = length(data) # Number of data

    xav = zeros(Float64, n+m)
    for i in 1:n
        xav[i] = xp[i]
    end
    for i in n+1:n+m
        xav[i] = data[i-n].mean
    end
    ccsq(x::Vector) = chisq(view(x, 1:n), view(x, n+1:n+m)) 
    if (n+m < 4)
        cfg = ForwardDiff.HessianConfig(ccsq, xav, ADerrors.Chunk{1}());
    else
        cfg = ForwardDiff.HessianConfig(ccsq, xav, ADerrors.Chunk{4}());
    end
        
    hess = Array{Float64}(undef, n+m, n+m)
    ForwardDiff.hessian!(hess, ccsq, xav, cfg)

    if (m-n > 0)
        m = length(data)
        n = size(hess, 1) - m
        
        Lm = LinearAlgebra.cholesky(LinearAlgebra.Symmetric(W))
        Li = LinearAlgebra.inv(Lm.L)
        
        hm = view(hess, 1:n, n+1:n+m)
        sm = hm * Li'
        
        maux = sm * sm'
        hi   = LinearAlgebra.pinv(maux)
        Px   = W - hm' * hi * hm
        
        
        
        nu = sqrt(C) * Px * sqrt(C)
        
        N = length(nu[1,:])
        z = randn(N, nmc)

        eig = abs.(eigvals(nu))
        eps = 1e-14 * maximum(eig)
        eig = eig .* (eig .> eps)

        aux = eig' * (z .^ 2)
        Q = 1.0 - juobs.mean(aux .< chi2)

        x = chi2 .- eig[2:end]' * (z[2:end,:].^2)
        x = x / eig[1]
        #dQ = juobs.mean((x .> 0) .* exp.(-x * 0.5) * 0.5 ./ sqrt.(abs.(x)))
        #dQ = err(cse)/value(cse) * dQ
    end
    return Q
end

function fve(mpi::uwreal, mk::uwreal, fpi::uwreal, fk::uwreal, ens::EnsInfo)
    mm = [6,12,8,6,24,24,0,12,30,24,24,8,24,48,0,6,48,36,24,24]
	nn = [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20]

    meta = sqrt(4/3 * mk ^ 2 - 1/3 * mpi ^ 2)

    jipi = value(mpi) ^ 2 / (4*pi*fpi) ^ 2
    jik = value(mk) ^ 2 / (4*pi*fpi) ^ 2
    jieta = value(meta) ^ 2 / (4*pi*fpi) ^ 2
    mpiL = value(mpi) * ens.L
    mkL = value(mk) * ens.L
    metaL = value(meta) * ens.L
    lampi = mpiL * sqrt.(nn)
    lamk = mkL * sqrt.(nn)
    lameta = metaL * sqrt.(nn)
    g1pi = sum(4 .* mm ./ lampi .* besselk.(1, lampi))
    g1k = sum(4 .* mm ./ lamk .* besselk.(1, lamk))
    g1eta = sum(4 .* mm ./ lameta .* besselk.(1, lameta))
    fve_mpi = 1/4 * jipi * g1pi - 1/12 * jieta * g1eta
    fve_mk = 1/6 * jieta * g1eta
    fve_fpi = - jipi * g1pi - 1/2 * jik * g1k
    fve_fk = -3/8 * jipi *g1pi - 3/4 * jik * g1k - 3/8 * jieta * g1eta

    mpi_infty = mpi / (1+fve_mpi)
    mk_infty = mk / (1+fve_mk)
    fpi_infty = fpi / (1+fve_fpi)
    fk_infty = fk / (1+fve_fk)

    return mpi_infty, mk_infty, fpi_infty, fk_infty
end

